/*
 * Copyright (C) 2001 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "screen.h"
#include "graphics.h"
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>

GtkWidget *combo2, *combo3;
GtkWidget *fill_button, *clip_button;
GtkWidget *gtklist;
GtkWidget *list_item;

/* Backing pixmap for drawing area */
static GdkPixmap *pixmap = NULL;
int zoom = 200;
int zoom_normal = 100;
int point_count = 0;
struct line_2d_t line, *selected_line = NULL;
struct poly_2d_t poly, *selected_poly = NULL;
enum line_alg_t { DDA, BRES } line_alg = DDA;
enum fill_alg_t { SEED, INV } fill_alg = SEED;
enum action_t { LINE, POLY, FILL, CLIP } action = LINE, prev_action = LINE;

void redraw_scene(GtkWidget *widget);
void gtklist_add_line(struct line_2d_t *line);
void gtklist_add_poly(struct poly_2d_t *poly);

/* Create a new backing pixmap of the appropriate size */
static gint
configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
  if (pixmap) gdk_pixmap_unref(pixmap);
  pixmap = gdk_pixmap_new(widget->window, widget->allocation.width,
                          widget->allocation.height, -1);
  redraw_scene(widget);

  return TRUE;
}

/* Redraw the screen from the backing pixmap */
static gint
expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_pixmap(widget->window,
                  widget->style->fg_gc[GTK_WIDGET_STATE(widget)], pixmap,
                  event->area.x, event->area.y, event->area.x, event->area.y,
                  event->area.width, event->area.height);

  return FALSE;
}

/* Clear the backing pixmap */
static void
clear_pixmap(GtkWidget *widget)
{
  gdk_draw_rectangle (pixmap, widget->style->black_gc, TRUE, 0, 0,
                      widget->allocation.width, widget->allocation.height);
}

void
project(int width, int height, int *x, int *y) // screen -> gtk
{
  *x = (*x * zoom) / zoom_normal;
  *y = (*y * zoom) / zoom_normal;
  *x = *x + ((width - ((SCREEN_WIDTH * zoom) / zoom_normal)) / 2);
  *y = *y + ((height - ((SCREEN_HEIGHT * zoom) / zoom_normal)) / 2);
}

void
unproject(int width, int height, int *x, int *y) // gtk -> screen
{
  *x = *x + ((((SCREEN_WIDTH * zoom) / zoom_normal) - width) / 2);
  *y = *y + ((((SCREEN_HEIGHT * zoom) / zoom_normal) - height) / 2);
  *x = (*x * zoom_normal) / zoom;
  *y = (*y * zoom_normal) / zoom;
}

/* Redraw the the scene */
void
redraw_scene(GtkWidget *widget)
{
  GdkRectangle update_rect;
  GdkGC *gc;
  GdkColor color;
  int i, j, x, y;
  long int col, old_col;
  unsigned char col_r, col_g, col_b;
  
  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = widget->allocation.width;
  update_rect.height = widget->allocation.height;

  clear_pixmap(widget);
  if (screen_buffer != NULL) {
    gc = gdk_gc_new(gtk_widget_get_parent_window(widget));
    old_col = -1;
    for(i = 0; i < update_rect.height; ++i) 
      for(j = 0; j < update_rect.width; ++j) {
	x = j; y = i;
	unproject(update_rect.width, update_rect.height, &x, &y);
	if (check_screen_coords(x, y)) {
	  col = screen_buffer[y][x];
	  if (col != old_col) {
	    old_col = col;
	    unrgb(col, &col_r, &col_g, &col_b);
	    color.red = col_r * 256; color.green = col_g * 256; 
	    color.blue = col_b * 256;
	    gdk_colormap_alloc_color(gdk_colormap_get_system(), &color, TRUE, 
				     TRUE);
	    gdk_gc_set_foreground (gc, &color);
	  }
	  gdk_draw_point(pixmap, gc, j, i);
	}
      }
  }

  gtk_widget_draw (widget, &update_rect);
}

static gint
button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  int x, y;
  struct point_2d_t p;
  
  if (event->button == 1) {
    x = event->x; y = event->y;
    unproject(widget->allocation.width, widget->allocation.height, &x, &y);
    if (check_screen_coords(x, y)) {
      switch (action) {
      case LINE : {
	line.points[point_count].x = x; line.points[point_count].y = y;
	point(line.points, rgb(0, 0, 0));
	point_count++;
	if (point_count > 1) {
	  point_count = 0;
	  gtklist_add_line(add_line(&line));
	  switch (line_alg) {
	  case DDA : { line_dda(&line, rgb(0, 0, 0), NULL); break; }
	  case BRES: line_bres(&line, rgb(0, 0, 0), NULL);
	  }
	}
	break;
      }
      case POLY : {
	p.x = x; p.y = y;
	poly_2d_add_point(&poly, &p);
	point(&p, rgb(0, 0, 0));
	point_count++;
	break;
      }
      case FILL : {
	p.x = x; p.y = y;
	switch (fill_alg) {
	case SEED : { fill_seed4(&p, rgb(0, 0, 0)); break; }
	default : {}
	}
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fill_button), FALSE);
	break;
      }
      case CLIP : {
	line.points[point_count].x = x; line.points[point_count].y = y;
	point_count++;
	if (point_count > 1) {
	  point_count = 0;
	  if (selected_line) {
	    switch (line_alg) {
	    case DDA : { 
	      line_dda(selected_line, rgb(255, 255, 255), NULL); break;
	    }
	    case BRES : line_bres(selected_line, rgb(255, 255, 255), NULL);
	    }
	    if (clip_line(selected_line, &line.points[0], &line.points[1])) {
	      switch (line_alg) {
	      case DDA : { 
		line_dda(selected_line, rgb(0, 128, 0), NULL); break;
	      }
	      case BRES : line_bres(selected_line, rgb(0, 128, 0), NULL);
	      }
	    }
	  }
	  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(clip_button), FALSE);
	}
      }
      }
    }
  }
  if (event->button == 3) {
    switch (action) {
    case POLY : {
      if (point_count > 1) { 
	point_count = 0;
	gtklist_add_poly(add_poly(&poly));
	poly_2d_draw(&poly, rgb(0, 0, 0));
	poly_2d_fini(&poly); poly_2d_init(&poly);
      }
    }
    default : {}
    }
  }

  redraw_scene(widget);
  
  return TRUE;
}

void
quit()
{
  gtk_exit(0);
}

void
combo1_changed(GtkWidget *widget, GtkEditable *entry)
{
  if (strcmp(gtk_entry_get_text(GTK_ENTRY(entry)), "line")) {
    action = POLY;
    gtk_widget_set_sensitive(combo2, FALSE);
    gtk_widget_set_sensitive(combo3, TRUE);
    gtk_widget_set_sensitive(fill_button, TRUE);
  }
  else {
    action = LINE;
    gtk_widget_set_sensitive(combo2, TRUE);
    gtk_widget_set_sensitive(combo3, FALSE);
    gtk_widget_set_sensitive(fill_button, FALSE);
  }
}

void
combo2_changed(GtkWidget *widget, GtkEditable *entry)
{
  if (strcmp(gtk_entry_get_text(GTK_ENTRY(entry)), "DDA")) 
    line_alg = BRES;
  else line_alg = DDA;
}

void
combo3_changed(GtkWidget *widget, GtkEditable *entry)
{
  if (strcmp(gtk_entry_get_text(GTK_ENTRY(entry)), "seed")) 
    fill_alg = INV;
  else fill_alg = SEED;
}

void
fill_button_toggled(GtkToggleButton *togglebutton, gpointer area)
{
  if (point_count == 0) {
    if (action != FILL) { 
      prev_action = action; action = FILL;
      if (fill_alg == INV) {
	if (selected_poly) {
	  fill_inv(selected_poly, rgb(0, 0, 0) ^ rgb(255, 255, 255));
	  redraw_scene(area);
	}
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fill_button), FALSE);
      }
    }
    else action = prev_action;
  }
  else gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fill_button), FALSE);
}

void
clip_button_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if (point_count == 0) {
    if (action != CLIP) { prev_action = action; action = CLIP; }
    else action = prev_action;
  }
  else gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(clip_button), FALSE);
}

void
set_zoom(GtkWidget *area, GtkEditable *entry)
{
  char *buffer;
  float fnum = 0;

  buffer = gtk_entry_get_text(GTK_ENTRY(entry));
  sscanf(buffer, "%f", &fnum);
  if (!fnum) gtk_entry_set_text(GTK_ENTRY(entry), "1");
  else { 
    zoom = (int)(fnum * 100);
    redraw_scene(area);
  }
}

void
gtklist_add_line(struct line_2d_t *line)
{
  GList *list = NULL;
  char buffer[20];

  sprintf(buffer, "line %i", line->id);
  list_item = gtk_list_item_new_with_label(buffer);
  list = g_list_append(list, list_item);
  gtk_widget_show(list_item);
  gtk_object_set_data(GTK_OBJECT(list_item), "line", line);
  gtk_list_append_items(GTK_LIST(gtklist), list);
}

void
gtklist_add_poly(struct poly_2d_t *poly)
{
  GList *list = NULL;
  char buffer[20];

  sprintf(buffer, "polygon %i", poly->id);
  list_item = gtk_list_item_new_with_label(buffer);
  list = g_list_append(list, list_item);
  gtk_widget_show(list_item);
  gtk_object_set_data(GTK_OBJECT(list_item), "poly", poly);
  gtk_list_append_items(GTK_LIST(gtklist), list);
}

void
gtklist_select(GtkList *list, GtkWidget *widget, gpointer area)
{
  if (selected_line) {
    switch (line_alg) {
    case DDA : { line_dda(selected_line, rgb(0, 0, 0), NULL); break; }
    case BRES: line_bres(selected_line, rgb(0, 0, 0), NULL);
    }
  }
  if (selected_poly) poly_2d_draw(selected_poly, rgb(0, 0, 0));
  selected_line = 
    ((struct line_2d_t *)gtk_object_get_data(GTK_OBJECT(widget), "line"));
  selected_poly = 
    ((struct poly_2d_t *)gtk_object_get_data(GTK_OBJECT(widget), "poly"));
  if (selected_line) {
    switch (line_alg) {
    case DDA : { line_dda(selected_line, rgb(255, 0, 0), NULL); break; }
    case BRES: line_bres(selected_line, rgb(255, 0, 0), NULL);
    }
  }
  if (selected_poly) poly_2d_draw(selected_poly, rgb(255, 0, 0));
  redraw_scene(area);
}

void
gtklist_unselect(GtkList *list, GtkWidget *widget, gpointer area)
{
  if (selected_line) {
    switch (line_alg) {
    case DDA : { line_dda(selected_line, rgb(0, 0, 0), NULL); break; }
    case BRES: line_bres(selected_line, rgb(0, 0, 0), NULL);
    }
    selected_line = NULL;
  }
  if (selected_poly) {
    poly_2d_draw(selected_poly, rgb(0, 0, 0));
    selected_poly = NULL;
  }
  redraw_scene(area);
}

void
painter_button_click(gpointer area, GtkButton *button)
{
  int n = 0, i = 0;
  long int c;
  struct poly_2d_t *p = first_poly;
  struct poly_3d_t **polys = NULL;

  while (p != NULL) { ++n; p = p->next; }
  polys = (struct poly_3d_t **)malloc(n * sizeof(struct poly_3d_t *));
  if (polys == NULL) return;
  p = first_poly;
  while (p != NULL) {
    polys[i] = NULL;
    polys[i] = (struct poly_3d_t *)malloc(sizeof(struct poly_3d_t));
    if (polys[i] == NULL) return;
    polys[i]->poly_2d = p;
    polys[i]->z = (int)(n*(rand()/(RAND_MAX+1.0)));
    ++i; p = p->next;
  }
  malir(polys, n);
  for (i = 0; i < n; ++i) {
    c = rgb((int)(255*(rand()/(RAND_MAX+1.0))),
	    (int)(255*(rand()/(RAND_MAX+1.0))),
	    (int)(255*(rand()/(RAND_MAX+1.0))));
    poly_2d_draw(polys[i]->poly_2d, c);
    //fill_inv(polys[i]->poly_2d, c);
  }
  redraw_scene(area);
}

int
main(int argc, char *argv[])
{
  GtkWidget *window;
  GtkWidget *drawing_area;
  GtkWidget *hbox, *vbox, *hbox_zoom, *hbox_line_alg, *hbox_fill_alg;
  GtkWidget *combo1;
  GList *combo_items = NULL;
  GtkWidget *entry;
  GtkWidget *zoom_label, *line_alg_label, *fill_alg_label;
  GtkWidget *quit_button, *painter_button;
  GtkWidget *scrolled_window;
  
  gtk_set_locale ();
  gtk_init (&argc, &argv);
  
  if (screen_buffer_init()) exit(1);
  screen_buffer_clear(rgb(255, 255, 255));

  poly_2d_init(&poly);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Graphics demo");

  gtk_signal_connect (GTK_OBJECT (window), "destroy",
                      GTK_SIGNAL_FUNC (quit), NULL);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), hbox);
  gtk_widget_show (hbox);
  
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_set_usize(scrolled_window, 100, 0);
  gtk_box_pack_start(GTK_BOX(hbox), scrolled_window, FALSE, FALSE, 0);
  gtk_widget_show(scrolled_window);

  gtklist = gtk_list_new();
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
					gtklist);
  gtk_widget_show(gtklist);

  /* Create the drawing area */
  
  drawing_area = gtk_drawing_area_new ();
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), 200, 200);
  gtk_box_pack_start (GTK_BOX (hbox), drawing_area, TRUE, TRUE, 0);
  gtk_widget_show (drawing_area);
  
    /* Signals used to handle backing pixmap */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
                      (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
                      (GtkSignalFunc) configure_event, NULL);

  /* Event signals */

  gtk_signal_connect (GTK_OBJECT (gtklist), "select_child",
                      (GtkSignalFunc) gtklist_select,
		      GTK_OBJECT (drawing_area));

  gtk_signal_connect (GTK_OBJECT (gtklist), "unselect_child",
                      (GtkSignalFunc) gtklist_unselect,
		      GTK_OBJECT (drawing_area));

  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
                      (GtkSignalFunc) button_press_event, NULL);

  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
                         | GDK_LEAVE_NOTIFY_MASK
                         | GDK_BUTTON_PRESS_MASK
                         | GDK_POINTER_MOTION_MASK
                         | GDK_POINTER_MOTION_HINT_MASK);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 0);
  gtk_widget_show (vbox);

  hbox_zoom = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox_zoom, TRUE, TRUE, 0);
  gtk_widget_show (hbox_zoom);

  zoom_label = gtk_label_new("Zoom: ");
  gtk_box_pack_start (GTK_BOX (hbox_zoom), zoom_label, FALSE, FALSE, 2);
  gtk_widget_show (zoom_label);
  
  entry = gtk_entry_new();
  gtk_box_pack_start (GTK_BOX (hbox_zoom), entry, TRUE, TRUE, 0);
  gtk_entry_set_text(GTK_ENTRY(entry), "2");
  gtk_widget_show (entry);

  gtk_signal_connect_object (GTK_OBJECT (entry), "activate",
                             (GtkSignalFunc) set_zoom,
                             GTK_OBJECT (drawing_area));

  combo1 = gtk_combo_new();
  combo_items = g_list_append(combo_items, "line");
  combo_items = g_list_append(combo_items, "polygon");
  gtk_combo_set_popdown_strings(GTK_COMBO(combo1), combo_items);
  gtk_combo_set_value_in_list(GTK_COMBO(combo1), TRUE, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), combo1, TRUE, TRUE, 0);
  gtk_widget_show (combo1);

  hbox_line_alg = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox_line_alg, TRUE, TRUE, 0);
  gtk_widget_show (hbox_line_alg);

  line_alg_label = gtk_label_new("Line alg.: ");
  gtk_box_pack_start (GTK_BOX (hbox_line_alg), line_alg_label, FALSE, FALSE,
		      2);
  gtk_widget_show (line_alg_label);

  combo2 = gtk_combo_new();
  combo_items = g_list_remove(combo_items, "line");
  combo_items = g_list_remove(combo_items, "polygon");
  combo_items = g_list_append(combo_items, "DDA");
  combo_items = g_list_append(combo_items, "Bresenham's");
  gtk_combo_set_popdown_strings(GTK_COMBO(combo2), combo_items);
  gtk_combo_set_value_in_list(GTK_COMBO(combo2), TRUE, FALSE);
  gtk_box_pack_start (GTK_BOX (hbox_line_alg), combo2, TRUE, TRUE, 0);
  gtk_widget_show (combo2);

  gtk_signal_connect_object (GTK_OBJECT(GTK_COMBO(combo1)->entry), "changed",
                             (GtkSignalFunc) combo1_changed, NULL);

  gtk_signal_connect_object (GTK_OBJECT(GTK_COMBO(combo2)->entry), "changed",
                             (GtkSignalFunc) combo2_changed, NULL);

  hbox_fill_alg = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox_fill_alg, TRUE, TRUE, 0);
  gtk_widget_show (hbox_fill_alg);

  fill_alg_label = gtk_label_new("Fill alg.: ");
  gtk_box_pack_start (GTK_BOX (hbox_fill_alg), fill_alg_label, FALSE, FALSE,
		      2);
  gtk_widget_show (fill_alg_label);

  combo3 = gtk_combo_new();
  combo_items = g_list_remove(combo_items, "DDA");
  combo_items = g_list_remove(combo_items, "Bresenham's");
  combo_items = g_list_append(combo_items, "seed");
  combo_items = g_list_append(combo_items, "inverse");
  gtk_combo_set_popdown_strings(GTK_COMBO(combo3), combo_items);
  gtk_combo_set_value_in_list(GTK_COMBO(combo3), TRUE, FALSE);
  gtk_box_pack_start (GTK_BOX (hbox_fill_alg), combo3, TRUE, TRUE, 0);
  gtk_widget_show (combo3);
  gtk_widget_set_sensitive(combo3, FALSE);

  gtk_signal_connect_object (GTK_OBJECT(GTK_COMBO(combo3)->entry), "changed",
                             (GtkSignalFunc) combo3_changed, NULL);

  fill_button = gtk_toggle_button_new_with_label("Fill polygon");
  gtk_box_pack_start (GTK_BOX (vbox), fill_button, TRUE, TRUE, 0);
  gtk_widget_show (fill_button);
  gtk_widget_set_sensitive(fill_button, FALSE);

  gtk_signal_connect (GTK_OBJECT (fill_button), "toggled",
                      (GtkSignalFunc) fill_button_toggled,
		      GTK_OBJECT (drawing_area));

  clip_button = gtk_toggle_button_new_with_label("Clip line");
  gtk_box_pack_start (GTK_BOX (vbox), clip_button, TRUE, TRUE, 0);
  gtk_widget_show (clip_button);

  gtk_signal_connect (GTK_OBJECT (clip_button), "toggled",
                      (GtkSignalFunc) clip_button_toggled, NULL);

  painter_button = gtk_button_new_with_label ("Painter");
  gtk_box_pack_start (GTK_BOX (vbox), painter_button, TRUE, TRUE, 0);
  gtk_widget_show (painter_button);
  
  gtk_signal_connect_object(GTK_OBJECT (painter_button), "clicked",
			    GTK_SIGNAL_FUNC (painter_button_click),
			    GTK_OBJECT (drawing_area));

  /* .. And a quit button */
  quit_button = gtk_button_new_with_label ("Quit");
  gtk_box_pack_start (GTK_BOX (vbox), quit_button, TRUE, TRUE, 0);
  gtk_signal_connect_object (GTK_OBJECT (quit_button), "clicked",
                             GTK_SIGNAL_FUNC (gtk_widget_destroy),
                             GTK_OBJECT (window));
  gtk_widget_show (quit_button);

  gtk_widget_show (window);
  
  gtk_main ();

  poly_2d_fini(&poly);

  screen_buffer_fini();  

  return 0;
}
